/*
 * Physically Based Simulation, SS 2005
 *
 * Markus Liechti (liechtim@student.ethz.ch)
 * Yoshimi Takano (yoshi@student.ethz.ch)
 * Stefan Wismer (swismer@student.ethz.ch)
 *
 * $Id: bridge.cpp,v 1.25 2005/06/25 20:00:04 ytakano Exp $
 */

#include <vector>
#include <stack>
#include <iostream>
#include <algorithm>
#include "bridge.h"

void Bridge::optimize(int level) {
	vector<bool> connectedVertex;
	connectedVertex.reserve(vertices.size());
	// initialize array
	for (unsigned int i = 0; i < vertices.size(); i++) {
		connectedVertex[i] = (vertices[i].type != vertex::normal);
	}
	// "activate" start and end vertex
	for (unsigned int i = 0; i < edges.size(); i++) {
		// check if start point and end point are the same
		if (edges[i].start != edges[i].end) {
			connectedVertex[edges[i].start] = true;
			connectedVertex[edges[i].end] = true;
		}
	}
	// delete "unactivated" vertices
	for (int i = vertices.size() - 1; i >= 0; i--) {
		if (!connectedVertex[i]) {
			// delete lonely vertex
			vertices.erase(vertices.begin() + i);
			// update start and end reference of edges array
			for (unsigned int j = 0; j < edges.size(); j++) {
				if (edges[j].start > i) edges[j].start--;
				if (edges[j].end > i) edges[j].end--;
			}
		}
	}

	if (level > 0) {
		connectedVertex.reserve(vertices.size());
		stack<int> toVisit;

		// initialize array and push fixed vertices to the stack
		for (unsigned int i = 0; i < vertices.size(); i++) {
			if (vertices[i].type != vertex::normal) {
				connectedVertex[i] = true;
				toVisit.push(i);
			} else {
				connectedVertex[i] = false;
			}
		}

		// pick a reachable vertex and add all neighbors
		while (!toVisit.empty()) {
			int cur = toVisit.top();
			toVisit.pop();
			connectedVertex[cur] = true;
			for (unsigned int i = 0; i < edges.size(); i++) {
				if (cur == edges[i].start && connectedVertex[edges[i].end] == false) {
					// we found a new vertex
					toVisit.push(edges[i].end);
				} else if (cur == edges[i].end && connectedVertex[edges[i].start] == false) {
					// we found a new vertex
					toVisit.push(edges[i].start);
				}
			}
		}

		// delete unused edges
		for (int i = edges.size() - 1; i >= 0; i--) {
			if (connectedVertex[edges[i].start] == false && connectedVertex[edges[i].end] == false) {
				cout << "nn" << i << endl;
				edges.erase(edges.begin() + i);
			}
		}

		// delete "unactivated" vertices
		for (int i = vertices.size() - 1; i >= 0; i--) {
			if (!connectedVertex[i]) {
				// delete lonely vertex
				vertices.erase(vertices.begin() + i);
				// update start and end reference of edges array
				for (unsigned int j = 0; j < edges.size(); j++) {
					if (edges[j].start > i) edges[j].start--;
					if (edges[j].end > i) edges[j].end--;
				}
			}
		}
	}
}

int Bridge::getNearestVertex(vector2 &pos) {
	component minDist = HUGE_VAL, tmp;
	int minIdx = -1;

	// compute distance to all vertices and pick nearest
	for (unsigned int i = 0; i < vertices.size(); i++) {
		tmp = length(pos - vertices[i].initialPosition);
		if (tmp < minDist) {
			minDist = tmp;
			minIdx = i;
		}
	}

	// if we found one and the distance is less than 1 then return this one, else return -1
	if (minIdx != -1 && minDist < 1) {
		return minIdx;
	} else {
		return -1;
	}
}

int Bridge::getNearestEdge(vector2 &pos) {
	component minDist = HUGE_VAL, d, var;
	vector2 u;
	int minIdx = -1;

	var = 1;
	// compute distance to all edges and pick nearest
	for (unsigned int i = 0; i < edges.size(); i++) {
		// check for line boundaries
		if (((vertices[edges[i].start].initialPosition.x - var <= pos.x && vertices[edges[i].end].initialPosition.x + var >= pos.x) ||
		(vertices[edges[i].start].initialPosition.x + var >= pos.x && vertices[edges[i].end].initialPosition.x - var <= pos.x)) &&
		((vertices[edges[i].start].initialPosition.y - var <= pos.y && vertices[edges[i].end].initialPosition.y + var >= pos.y) ||
		(vertices[edges[i].start].initialPosition.y + var >= pos.y && vertices[edges[i].end].initialPosition.y - var <= pos.y))) {
			// formula: d = (|(q-p) x u| / |u|)
			u = vertices[edges[i].end].initialPosition - vertices[edges[i].start].initialPosition;
			d = abs((pos - vertices[edges[i].start].initialPosition) % u) / length(u);
			if (d <= minDist) {
				minDist = d;
				minIdx = i;
			}
		}
	}
	// if we found one and the distance is less than 1 then return this one, else return -1
	if (minIdx != -1 && minDist < 1) {
		return minIdx;
	} else {
		return -1;
	}
}

bool Bridge::getRoad() {
	road.clear();
	component left = HUGE_VAL, right = 0.0, height = 0.0, tmp, check;

	// find left and right road boundary
	for (unsigned i = 0; i < vertices.size(); i++) {
		if (vertices[i].type == vertex::fixedTrainStartOrEnd) {
			if (vertices[i].initialPosition.x < left) {
				right = left;
				roadEndVertex = i;
				left = vertices[i].initialPosition.x;
				roadStartVertex = i;
				height = vertices[i].initialPosition.y;
			} else {
				right = vertices[i].initialPosition.x;
				roadEndVertex = i;
			}
		}
	}

	for (unsigned i = 0; i < edges.size(); i++) edges[i].isRoad = false;

	// get indices of all road components (edges) between left and the right boundary
	tmp = left;
	while (tmp != right) {
		check = tmp;
		for (unsigned i = 0; i < edges.size(); i++) {
			// check if edge is horizontal and has the right height
			if (vertices[edges[i].start].initialPosition.y == height &&
				vertices[edges[i].end].initialPosition.y == height) {
				if (vertices[edges[i].start].initialPosition.x == tmp &&
					vertices[edges[i].start].initialPosition.x < vertices[edges[i].end].initialPosition.x) {
					tmp =  vertices[edges[i].end].initialPosition.x;
					edges[i].isRoad = true;
					road.push_back(i);
				}
				if (vertices[edges[i].end].initialPosition.x == tmp &&
					vertices[edges[i].end].initialPosition.x < vertices[edges[i].start].initialPosition.x) {
					tmp =  vertices[edges[i].start].initialPosition.x;
					edges[i].isRoad = true;
					road.push_back(i);
				}
			}
		}
		// check if a new road part has been found, if not clear road vector
		if (tmp == check) {
			for (unsigned int i = 0; i < edges.size(); i++) edges[i].isRoad = false;
			road.clear();
			return false;
		}
	}
	return true;
}

int Bridge::addVertex(int x, int y) {
	vertex tmp;
	tmp.initialPosition.x = x; tmp.initialPosition.y = y;
	return addVertex(tmp);
}

int Bridge::addVertex(vertex v) {
	// is a vertex with the same coordinates already present in the vector?
	vector<vertex>::iterator i = find(vertices.begin(), vertices.end(), v);
	if (i == vertices.end()) { // vertex not in vector
		v.type = vertex::normal;
		vertices.push_back(v);
		return vertices.size() - 1;
	} else {
		return i - vertices.begin(); // position in vector
	}
}

int Bridge::addEdge(int start, int end) {
	edge tmp;
	tmp.start = start; tmp.end = end;

	// check if start and end references are valid
	if ((unsigned)start < vertices.size() && (unsigned)end < vertices.size()) {
		// is an edge with the same start and end already present in the vector?
		vector<edge>::iterator i = find(edges.begin(), edges.end(), tmp);
		if (i == edges.end()) { // edge not in vector
			edges.push_back(tmp);
			return edges.size() - 1;
		} else {
			return i - edges.begin(); // position in vector
		}
	} else {
		return -1;
	}
}

void Bridge::deleteVertex(int index) {
	if ((unsigned)index >= 0 && (unsigned)index < vertices.size()) {
		if (vertices[index].type != vertex::normal) return;

		// delete adjacent edges
		for (int i = edges.size() - 1; i >= 0; i--) {
			if (edges[i].start == index) {
				edges.erase(edges.begin() + i);
			} else {
				if (edges[i].start > index) edges[i].start--;
				if (edges[i].end == index) {
					edges.erase(edges.begin() + i);
				} else {
					if (edges[i].end > index) edges[i].end--;
				}
			}
		}
		// delete vertex itself
		vertices.erase(vertices.begin() + index);
		optimize(0);
	}
}

void Bridge::deleteEdge(int index) {
	if ((unsigned)index >= 0 && (unsigned)index < edges.size()) {
		edges.erase(edges.begin() + index);
	}
	optimize(0);
}

void Bridge::clear() {
	edges.clear();
	// clear all normal (except fixed) vertices
	for (int i = vertices.size() - 1; i >= 0; i--) {
		if (vertices[i].type == vertex::normal) {
			vertices.erase(vertices.begin() + i);
		}
	}
}

vector2 &Bridge::getStart(int index) {
	return vertices[edges[index].start].position;
}

vector2 &Bridge::getEnd(int index) {
	return vertices[edges[index].end].position;
}

void Bridge::initializePhysics() {
	optimize(1);
	
	for (unsigned int i = 0; i < vertices.size(); i++) {
		vertices[i].velocity = vector2(0.0, 0.0);
		vertices[i].position = vertices[i].initialPosition;
		vertices[i].mass = 0;
	}

	for (unsigned int i = 0; i < edges.size(); i++) {
		edges[i].isBroken = false;

		// compute length and masses
		vertex &start = vertices[edges[i].start];
		vertex &end = vertices[edges[i].end];
		component l = length(end.initialPosition - start.initialPosition);
		edges[i].length = edges[i].initialLength = l;
		start.mass += l/2.0 * EDGE_DENSITY;
		end.mass += l/2.0 * EDGE_DENSITY;
	}
}

void Bridge::setExternalForces(vector<vertex> &v) {
	vector2 gravityForce(0.0, -GRAVITY); 
	for (unsigned i = 0; i < v.size(); i++) {
		v[i].force = vertices[i].mass * gravityForce;
		if (v[i].forceCoach != 0) {
			v[i].force.y -= v[i].forceCoach;
		}
	}
}

void Bridge::addInternalForces(vector<vertex> &v, vector<edge> &e, component stiffness, bool breakableEdges = true) {
	for (unsigned i = 0; i < e.size(); i++) {
		if (!e[i].isBroken) {
			vertex &start = v[e[i].start];
			vertex &end = v[e[i].end];
			component l = length(end.position - start.position);
			if (l == 0) {
				l = 0.01;
			}

			vector2 elasticForce = stiffness * (l - e[i].initialLength) * (end.position - start.position) / l;

			e[i].length = l;

			component deformation = e[i].deformation();
			
			if (breakableEdges && (deformation > EDGE_TENSION_THRESHOLD || deformation < EDGE_PRESSURE_THRESHOLD) ) {
				e[i].isBroken = true;
				puff.createPuff(start.position, end.position);
			}

			start.force += elasticForce;
			end.force -= elasticForce;
		}
	}
}

void Bridge::doPhysics(component timestep) {
	setExternalForces(vertices);
	
	addInternalForces(vertices, edges, EDGE_STIFFNESS);
	
	// leap frog integration
	for (unsigned int i = 0; i < vertices.size(); i++) {
		if (vertices[i].type >= vertex::fixed) {
			continue;
		}

		vector2 a = (vertices[i].force - DAMPING * vertices[i].velocity - DAMPING_SQUARE * signSquare(vertices[i].velocity)) / vertices[i].mass;

		vertices[i].velocity += timestep * a;
		vertices[i].position += timestep * vertices[i].velocity;
	}
	
	// perform puff simulations
	puff.doPhysics(timestep);
}
