/*
 * Physically Based Simulation, SS 2005
 *
 * Markus Liechti (liechtim@student.ethz.ch)
 * Yoshimi Takano (yoshi@student.ethz.ch)
 * Stefan Wismer (swismer@student.ethz.ch)
 *
 * $Id: bridge.h,v 1.23 2005/06/25 20:00:04 ytakano Exp $
 */
 
/** @file */

#ifndef __bridge_H__
#define __bridge_H__

#include <vector>
#include <cmath>
#include "structures.h"
#include "smoke.h"
#include "puff.h"

/** The linear damping coefficient. */
#define DAMPING 4000.0

/** The quadratic coefficient. */
#define DAMPING_SQUARE 3000.0

/** The gravity acceleration. */
#define GRAVITY 9.81

/** The edge density. */
#define EDGE_DENSITY 1200.0

/** The edge stiffness. */
#define EDGE_STIFFNESS 10000000.0

/** The edge tension threshold. */
#define EDGE_TENSION_THRESHOLD 0.038

/** The edge pressure threshold. */
#define EDGE_PRESSURE_THRESHOLD -0.050

using namespace std;

/**
 * Represents a bridge.
 */
class Bridge {
public:
	/** The bridge vertices. */
	vector<vertex> vertices;

	/** The bridge edges. */
	vector<edge> edges;

	/** The "puff" particle system (to animate broken edges). */
	Puff puff;
	
	/*
	 * static methods
	 */
	
	/**
	 * Removes unnecessary vertices and edges.
	 * - level 0: deletes vertices that are no longer connected
	 * - level 1: deletes vertices and edges that are not reachable from the
	 * ground ('flying' elements)
	 * @param level the desired optimization level
	 */
	void optimize(int level);

	/**
	 * Finds the index of the nearest vertex as to a given position.
	 * @param pos the position
	 * @return the index of the nearest vertex or -1 if the distance is greater
	 * than 1
	 */
	int getNearestVertex(vector2 &pos);
	
	/**
	 * Finds the index of the nearest edge as to a given position.
	 * @param pos the position
	 * @return the index of the nearest edge or -1 if the distance is 'too large'
	 */
	int getNearestEdge(vector2 &pos);
	
	/**
	 * Fills the road array (indices of the edges representing the horizontal
	 * end-to-end road from left to right).
	 * @return true iff valid road can be found
	 */
	bool getRoad();
	
	/**
	 * Adds a vertex to the bridge. Has no effect if the same vertex already
	 * exists.
	 * @param x the x coordinate of the vertex to be added
	 * @param y the y coordinate of the vertex to be added
	 * @return the index of the newly added (or already present) vertex
	 * @sa addVertex(vertex)
	 */
	int addVertex(int x, int y);
	
	/**
	 * Adds a vertex to the bridge. Has no effect if the same vertex already
	 * exists.
	 * @param v the vertex to be added
	 * @return the index of the newly added (or already present) vertex
	 * @sa addVertex(int, int)
	 */
	int addVertex(vertex v);
	
	/**
	 * Adds an edge to the bridge. Has no effect if the same edge already
	 * exists.
	 * @param start the index of the edge's start vertex
	 * @param end the index of the end's start vertex
	 * @return the index of the newly added (or already present) edge
	 */
	int addEdge(int start, int end);
	
	/**
	 * Deletes a vertex as well as all adjacent edges from the bridge.
	 * Afterwards, a call to @c optimize(0) is performed.
	 * @param index the index of the vertex to be deleted
	 * @sa optimize(int)
	 */
	void deleteVertex(int index);
	
	/**
	 * Deletes an edge from the bridge. Afterwards, a call to @c optimize(0) is
	 * performed.
	 * @param index the index of the edge to be deleted
	 * @sa optimize(int)
	 */
	void deleteEdge(int index);
	
	/** Clears the bridge (removes all vertices and edges). */
	void clear();
	
	/**
	 * Returns the position of the start vertex of edges[index]
	 * @param index the index of the edge
	 */
	vector2 &getStart(int index);
	
	/**
	 * Returns the position of the end vertex of edges[index]
	 * @param index the index of the edge
	 */
	vector2 &getEnd(int index);

	/**
	 * The indices of the edges composing the horizontal end-to-end road from
	 * left to right
	 */
	vector<int> road;
	
	/** The index of the road's start vertex. */
	int roadStartVertex;
	
	/** The index of the road's end vertex. */
	int roadEndVertex;
	
	/*
	 * dynamic methods
	 */

	/**
	 * Sets external forces (gravity etc.) for vertices in a given vector.
	 * @param v the vector of vertices to be processed
	 */
	void setExternalForces(vector<vertex> &v);
	
	/**
	 * Adds internal spring forces to vertices in a given vector.
	 * @param v the vector of vertices to be processed
	 * @param e the vector of edges to be processed
	 * @param stiffness the spring stiffness
	 * @param breakableEdges true iff the edges should be marked as broken when
	 * their relative deformation exceeds the deformation threshold
	 */
	void addInternalForces(vector<vertex> &v, vector<edge> &e, component stiffness, bool breakableEdges);
	
public:
	
	/**
	 * Initializes the bridge simulation. Should be called before
	 * doPhysics(component).
	 */
	void initializePhysics();
	
	/**
	 * Performs a step in the bridge simulation.
	 * @param timestep the time step
	 */
	void doPhysics(component timestep);
	
};

#endif
