/*
 * Physically Based Simulation, SS 2005
 *
 * Markus Liechti (liechtim@student.ethz.ch)
 * Yoshimi Takano (yoshi@student.ethz.ch)
 * Stefan Wismer (swismer@student.ethz.ch)
 *
 * $Id: bridgeGUI.cpp,v 1.28 2005/07/05 18:29:09 ytakano Exp $
 */

#include <string>
#include <vector>
#include "bridgeGUI.h"

using namespace std;

void BridgeGUI::initTextures() {
	gltexture = new GLuint[2];
	glGenTextures(2, &gltexture[0]);

	glBindTexture(GL_TEXTURE_2D, gltexture[0]);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D, 0, 4, world->train.textureWidth, world->train.textureHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, world->train.image);

	glBindTexture(GL_TEXTURE_2D, gltexture[1]);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D, 0, 4, world->train.smoke.textureWidth, world->train.smoke.textureHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, world->train.smoke.image);

	glBindTexture(GL_TEXTURE_2D, 0);
}

void BridgeGUI::setHSVcolor(float h, float s, float v) {
	// from the GDV1 lecture notes
	if (s == 0.0) {
		if (h == -1.0)
			glColor3f(v, v, v);
		else
			throw -1;
	} else {
		if (h == 360.0) h = 0.0;
		h /= 60.0;
		float i = floor(h);
		float f = h - i;
		float m = v * (1.0 - s);
		float n = v * (1.0 - s * f);
		float k = v * (1.0 - s * (1.0 - f));
		if (i == 0.0) { glColor3f(v, k, m); } 
		if (i == 1.0) { glColor3f(n, v, m); } 
		if (i == 2.0) { glColor3f(m, v, k); } 
		if (i == 3.0) { glColor3f(m, n, v); } 
		if (i == 4.0) { glColor3f(k, m, v); } 
		if (i == 5.0) { glColor3f(v, m, n); }
	}
}

vector2 BridgeGUI::getPos(int x, int y) {
	return vector2(x * viewSize.x / windowWidth + viewPos.x,
		viewSize.y - y * viewSize.y / windowHeight + viewPos.y);
}

vector2 BridgeGUI::roundPos(const vector2 &v) {
	return vector2(floor(v.x + 0.5), floor(v.y + 0.5));
}

void BridgeGUI::doMouseDetection(int x, int y) {
	cursor = getPos(x, y);
	selVertex = world->bridge.getNearestVertex(cursor);
	if (selVertex != -1) {
		cursor = world->bridge.vertices[selVertex];
	} else {
		selEdge = world->bridge.getNearestEdge(cursor);
	}
	cursor = roundPos(cursor);
}

void BridgeGUI::initGL(int Width, int Height) {
	glViewport(0, 0, Width, Height);
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClearDepth(1.0);
	glDepthFunc(GL_LEQUAL); // type of depth test
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_TEXTURE_2D);
	glShadeModel(GL_SMOOTH); // smooth color shading

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity(); // reset the projection matrix

	glOrtho(viewPos.x, viewPos.x + viewSize.x, viewPos.y, viewPos.y + viewSize.y, -4.0, 4.0);
	glMatrixMode(GL_MODELVIEW);
}

void BridgeGUI::drawGLScene() {
	Bridge &bridge = world->bridge;

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();

	// draw horizon
	glColor3f(0.4, 0.4, 0.4);
	glBegin(GL_QUAD_STRIP);
	for (vector<vertex>::iterator i = world->horizon.begin(); i != world->horizon.end(); i++) {
		glVertex3f(i->initialPosition.x, i->initialPosition.y, 1.0);
		glVertex3f(i->initialPosition.x, 0.0, 1.0);
	}
	glEnd();

	if (mode <= designMode) {
		// draw raster
		glColor3f (0.3, 0.3, 0.3);
		glBegin(GL_LINES);
		// vertical lines
		for (int i = 0; i < viewSize.x; i++) {
			glVertex3f(viewPos.x + i, viewPos.y, 0.0);
			glVertex3f(viewPos.x + i, viewPos.y + viewSize.y, 0.0);
		}
		// horizontal lines
		for (int i = 0; i < viewSize.y; i++) {
			glVertex3f(viewPos.x, viewPos.y + i, 0.0);
			glVertex3f(viewPos.x + viewSize.x, viewPos.y + i, 0.0);
		}
		glEnd();

		// draw vertices
		glBegin(GL_QUADS);
		for (vector<vertex>::iterator i = bridge.vertices.begin(); i != bridge.vertices.end(); i++) {
			component x = i->initialPosition.x, y = i->initialPosition.y;
			if (i->type >= vertex::fixed) {
				glColor3f(0.0, 0.8, 0.0);
			} else {
				glColor3f(1.0, 1.0, 1.0);
			}
			glVertex3f(x - 0.2, y + 0.2, 2.0);
			glVertex3f(x + 0.2, y + 0.2, 2.0);
			glVertex3f(x + 0.2, y - 0.2, 2.0);
			glVertex3f(x - 0.2, y - 0.2, 2.0);
		}

		// highlight selected vertex
		if (selVertex != -1) {
			component x = bridge.vertices[selVertex].initialPosition.x;
			component y = bridge.vertices[selVertex].initialPosition.y;
			glColor3f (1.0, 1.0, 0.0);
			glVertex3f(x - 0.2, y + 0.2, 2.1);
			glVertex3f(x + 0.2, y + 0.2, 2.1);
			glVertex3f(x + 0.2, y - 0.2, 2.1);
			glVertex3f(x - 0.2, y - 0.2, 2.1);
		}
		glEnd();

		// draw edges
		glEnable(GL_LINE_SMOOTH);
		glLineWidth(2.5);
		glColor3f(1.0, 1.0, 1.0);

		glBegin(GL_LINES);
		for (vector<edge>::iterator i = bridge.edges.begin(); i != bridge.edges.end(); i++) {
			glVertex3f(bridge.vertices[i->start].initialPosition.x, bridge.vertices[i->start].initialPosition.y, 1.0);
			glVertex3f(bridge.vertices[i->end].initialPosition.x, bridge.vertices[i->end].initialPosition.y, 1.0);
		}

		// highlight selected edge
		if (selEdge != -1) {
			glColor3f(1.0, 1.0, 0.0);
			glVertex3f(bridge.vertices[bridge.edges[selEdge].start].initialPosition.x, bridge.vertices[bridge.edges[selEdge].start].initialPosition.y, 1.1);
			glVertex3f(bridge.vertices[bridge.edges[selEdge].end].initialPosition.x, bridge.vertices[bridge.edges[selEdge].end].initialPosition.y, 1.1);
			glColor3f(1.0, 1.0, 1.0);	
		}

		// draw unfinished edge to cursor position
		if (drawVertex != -1) {
			if (length(bridge.vertices[drawVertex].initialPosition - cursor) > EDGE_MAX_LENGTH) glColor3f(1.0, 0.0, 0.0);
			glVertex3f(bridge.vertices[drawVertex].initialPosition.x, bridge.vertices[drawVertex].initialPosition.y, 1.0);
			glVertex3f(cursor.x, cursor.y, 1.0);	
		}
		glEnd();
		glDisable(GL_LINE_SMOOTH);
		glLineWidth(1.0);

		// draw cursor
		glBegin(GL_QUADS);
		glColor3f(0.0, 0.0, 1.0);
		glVertex3f(cursor.x - 0.2, cursor.y + 0.2, 3.0);
		glVertex3f(cursor.x + 0.2, cursor.y + 0.2, 3.0);
		glVertex3f(cursor.x + 0.2, cursor.y - 0.2, 3.0);
		glVertex3f(cursor.x - 0.2, cursor.y - 0.2, 3.0);
		glColor3f(1.0, 1.0, 1.0);
		glEnd();
	} else { // simulation is running
		// draw vertices
		glBegin(GL_QUADS);
		for (vector<vertex>::iterator i = bridge.vertices.begin(); i != bridge.vertices.end(); i++) {
			component x = i->position.x, y = i->position.y;
			if (i->type >= vertex::fixed) {
				glColor3f(0.0, 0.8, 0.0);
			} else {
				glColor3f(1.0, 1.0, 1.0);
			}
			glVertex3f(x - 0.2, y + 0.2, 2.0);
			glVertex3f(x + 0.2, y + 0.2, 2.0);
			glVertex3f(x + 0.2, y - 0.2, 2.0);
			glVertex3f(x - 0.2, y - 0.2, 2.0);
		}
		glEnd();

		// draw edges
		glEnable(GL_LINE_SMOOTH);
		glLineWidth(2.5);
		glBegin(GL_LINES);
		for (vector<edge>::iterator i = bridge.edges.begin(); i != bridge.edges.end(); i++) {
			if (!i->isBroken) {
				component deformation = i->deformation();
				float hue;
				if (deformation >= 0) {
					// use hue from green (120.0) to red (0.0)
					hue = -120.0f / EDGE_TENSION_THRESHOLD * i->deformation() + 120.0f;
				} else {
					// use hue from green (120.0) to blue (240.0)
					hue = 120.0f / EDGE_PRESSURE_THRESHOLD * i->deformation() + 120.0f;
				}
				setHSVcolor(hue, 1.0f, 0.8f);
				glVertex3f(bridge.vertices[i->start].position.x, bridge.vertices[i->start].position.y, 1.0);
				glVertex3f(bridge.vertices[i->end].position.x, bridge.vertices[i->end].position.y, 1.0);
			}
		}
		glEnd();
		glDisable(GL_LINE_SMOOTH);
		glLineWidth(1.0);

		if (mode == trainMode) {
			// draw train
			glBindTexture(GL_TEXTURE_2D, gltexture[0]);
			world->train.draw();
			glBindTexture(GL_TEXTURE_2D, 0);
		}

		// draw smoke and puff particles
		glBindTexture(GL_TEXTURE_2D, gltexture[1]);
		world->train.smoke.draw();
		glBindTexture(GL_TEXTURE_2D, 0);
		bridge.puff.draw();
	}

	// swap buffers to display, since we're double buffered
	SDL_GL_SwapBuffers();
}

int BridgeGUI::handleEvents() {
	SDL_Event event;
	while (SDL_PollEvent(&event)) {
		switch (event.type) {
			case SDL_QUIT:
				return 1;
				break;

			case SDL_KEYDOWN:
				switch (event.key.keysym.sym) {
					case SDLK_ESCAPE:
						return 1;
						break;
					case SDLK_1:
						if (event.key.keysym.mod & KMOD_CTRL)
							world->saveToFile(SAVE_BRIDGE_1);
						else
							world->loadFromFile(SAVE_BRIDGE_1);
						break;
					case SDLK_2:
						if (event.key.keysym.mod & KMOD_CTRL)
							world->saveToFile(SAVE_BRIDGE_2);
						else
							world->loadFromFile(SAVE_BRIDGE_2);
						break;
					case SDLK_3:
						if (event.key.keysym.mod & KMOD_CTRL)
							world->saveToFile(SAVE_BRIDGE_3);
						else
							world->loadFromFile(SAVE_BRIDGE_3);
						break;
					case SDLK_4:
						if (event.key.keysym.mod & KMOD_CTRL)
							world->saveToFile(SAVE_BRIDGE_4);
						else
							world->loadFromFile(SAVE_BRIDGE_4);
						break;
					case SDLK_c: // clear bridge
						world->bridge.clear();
						break;
					case SDLK_s: // start/stop simulation
						switch (mode) {
							case initMode:
							case designMode:
								world->bridge.initializePhysics();
								world->bridge.getRoad();
								mode = simMode;
								break;
							case simMode:
								world->train.smoke.clear();
								world->bridge.puff.clear();
								mode = designMode;
								break;
							case trainMode:
								world->train.finalize();
								world->train.smoke.clear();
								mode = simMode; // return to simulation mode
						}
						break;
					case SDLK_t: // start/stop train
						switch (mode) {
							case initMode:
							case designMode:
								world->bridge.initializePhysics();
								world->bridge.getRoad();
								mode = simMode;
								// no break!
							case simMode:
								if (world->train.initialize(world->bridge))
									mode = trainMode; // change mode if init OK
								break;
							case trainMode:
								world->train.finalize();
								world->train.smoke.clear();
								mode = simMode; //return to simulation mode
						}
						break;
					default:
						break;
				}
				break;

			case SDL_MOUSEMOTION:
				if (mode == initMode) mode = designMode;
				doMouseDetection(event.motion.x, event.motion.y);
				break;

			case SDL_MOUSEBUTTONDOWN:
				if (mode != designMode) break;
				if (event.button.button == SDL_BUTTON_LEFT) {
					// insert start-vertex; use cursor detected in MOUSEMOTION
					drawVertex = world->bridge.addVertex(cursor);
				}
				if (event.button.button == SDL_BUTTON_RIGHT) {
					// detect the position again for precision
					cursor = getPos(event.motion.x, event.motion.y);
					// delete vertex or edge
					int idxVertex = world->bridge.getNearestVertex(cursor);
					if (idxVertex != -1) {
						// vertex found: delete it
						world->bridge.deleteVertex(idxVertex);
						selVertex = -1;
					} else {
						// no vertex found: search for edge
						int idxEdge = world->bridge.getNearestEdge(cursor);
						// edge found: delete it
						if (idxEdge != -1) {
							world->bridge.deleteEdge(idxEdge);
							selEdge = -1;
						}
					}
					selEdge = selVertex = -1;
					cursor = roundPos(cursor);
				}
				break;

			case SDL_MOUSEBUTTONUP:
				if (mode != designMode) break;
				if (event.button.button == SDL_BUTTON_LEFT) {
					if (length(world->bridge.vertices[drawVertex].initialPosition - cursor) <= EDGE_MAX_LENGTH) {
						// insert end-vertex and edge; use cursor detected in MOUSEMOTION
						int endVertex = world->bridge.addVertex(cursor);
						world->bridge.addEdge(drawVertex, endVertex);
					} else {
						// remove drawVertex if necessary
						world->bridge.optimize(0);
					}
					drawVertex = -1;
				}
				break;
			default:
				break;
		}
	}
	int x,y;
	SDL_GetMouseState(&x, &y);
	if (mode != initMode && SDL_GetTicks() - lastScrollTime > 50) {// avoid scrolling when application has just started
		if (x < scrollBorder && viewPos.x > viewLeftBound) {
			viewPos.x--;
			initGL(windowWidth, windowHeight);
			doMouseDetection(x, y);
			lastScrollTime = SDL_GetTicks();
		}
		if (x > windowWidth - scrollBorder && viewPos.x < viewRightBound) {
			viewPos.x++;
			initGL(windowWidth, windowHeight);
			doMouseDetection(x, y);
			lastScrollTime = SDL_GetTicks();
		}
	}
	return 0;
}

void BridgeGUI::run(int width, int height, bool fullscreen, char *bridgeFile, char *trainFile) {
	windowWidth = width; windowHeight = height;
	viewPos.x = 0; viewPos.y = 0;
	viewSize.x = 64; viewSize.y = 48; // number of raster lines visible on the screen
	scrollBorder = 10;
	lastScrollTime = 0;

	cursor.x = 0; cursor.y = 0;
	drawVertex = -1;
	selVertex = -1; selEdge = -1;

	// load bridge from file
	bool bridgeLoaded = false;
	if (bridgeFile != NULL) {
		if (!world->loadFromFile(bridgeFile)) {
			cerr << "Unable to open bridge file: " << bridgeFile;
			cerr << ", using default bridge instead" << endl;
		} else {
			bridgeLoaded = true;
		}
	}

	if (!bridgeLoaded) {
		if (!world->loadFromFile(DEFAULT_BRIDGE)) {
			cerr << "Unable to open default bridge " << DEFAULT_BRIDGE << endl;
			exit(1);
		}
	}

	viewLeftBound = (int) world->horizon[0].initialPosition.x;
	viewRightBound = (int) world->horizon[world->horizon.size() - 1].initialPosition.x - (int) viewSize.x;

	// load train from file
	bool trainLoaded = false;
	if (trainFile != NULL) {
		if (!world->train.loadFromFile(trainFile)) {
			cerr << "Unable to open train file: " << trainFile;
			cerr << ", using default train instead" << endl;
		} else {
			trainLoaded = true;
		}
	}

	if (!trainLoaded) {
		if (!world->train.loadFromFile(DEFAULT_TRAIN)) {
			cerr << "Unable to open default train " << DEFAULT_TRAIN << endl;
			exit(1);
		}
	}

	// initialize SDL for video output
	if (SDL_Init(SDL_INIT_VIDEO) < 0) {
		cerr << "Unable to initialize SDL: " << SDL_GetError() << endl;
		exit(1);
	}

	Uint32 flags = SDL_OPENGL;
	if (fullscreen) flags |= SDL_FULLSCREEN;
	// create an OpenGL screen
	if (SDL_SetVideoMode(windowWidth, windowHeight, 0, flags) == NULL) {
		cerr << "Unable to create OpenGL screen: " << SDL_GetError() << endl;
		SDL_Quit();
		exit(2);
	}
	initTextures();

	// set the title bar
	SDL_WM_SetCaption("S12 - Bridge Simulator", NULL);

	cout << "S12 - Bridge Simulator 0.1" << endl;
	cout << "press Ctrl+[1..4] to save brigde, [1..4] to load" << endl;
	cout << "press c to clear bridge" << endl;
	cout << "press s to start/stop simulation, press t to start/stop train" << endl;
	cout << "press escape to quit" << endl;

	// loop, drawing and checking events
	initGL(windowWidth, windowHeight);
	mode = initMode;
	int done = 0;
	Uint32 lastTick = SDL_GetTicks(), deltaT;
	while (!done) {
		done = handleEvents();

		deltaT = SDL_GetTicks() - lastTick;
		lastTick = SDL_GetTicks();
		if (deltaT == 0) deltaT = 1; // avoid division by zero
		if (deltaT > 5) deltaT = 5; // make it more stable

		component timestep = static_cast<component>(deltaT) / 1000.0; // seconds
		if (mode < simMode) {
			SDL_Delay(20);
		} else if (mode == simMode) {
			world->bridge.doPhysics(timestep);
		} else if (mode == trainMode) {
			world->bridge.doPhysics(timestep);
			if (world->train.doPhysics(timestep)) // train passed bridge
				mode = simMode;
		}

		drawGLScene();
	}
	SDL_Quit();
}
