/*
 * Physically Based Simulation, SS 2005
 *
 * Markus Liechti (liechtim@student.ethz.ch)
 * Yoshimi Takano (yoshi@student.ethz.ch)
 * Stefan Wismer (swismer@student.ethz.ch)
 *
 * $Id: bridgeGUI.h,v 1.20 2005/06/25 20:00:04 ytakano Exp $
 */
 
 /** @file */

#ifndef __bridgeGUI_H__
#define __bridgeGUI_H__

#include "opengl.h"
#include "structures.h"
#include "world.h"
#include "bridge.h"

/** The maximum edge length. */
#define EDGE_MAX_LENGTH 9.0

/** The path to the default bridge description file. */
#define DEFAULT_BRIDGE "data/example.brg"

/** The path to the default train description file. */
#define DEFAULT_TRAIN "data/sbahn.trn"

/** The path to the saved bridge description file 1. */
#define SAVE_BRIDGE_1 "data/save1.brg"

/** The path to the saved bridge description file 2. */
#define SAVE_BRIDGE_2 "data/save2.brg"

/** The path to the saved bridge description file 3. */
#define SAVE_BRIDGE_3 "data/save3.brg"

/** The path to the saved bridge description file 4. */
#define SAVE_BRIDGE_4 "data/save4.brg"

/**
 * The graphical user interface.
 */
class BridgeGUI {
private:
	World *world;                      // a pointer to the world for simpler access
	int windowWidth, windowHeight;     // width and height of the window
	int scrollBorder;                  // size of scroll-border
	int lastScrollTime;                // time of last scolling action. used for limiting scrolling rate

	vector2 viewPos, viewSize;         // used for specifying which area to draw
	int viewLeftBound, viewRightBound; // bounds of the world, scrolling limits in the GUI
	vector2 cursor;                    // cursor in world coordinates

	int selVertex, selEdge;            // index of selected vertex / edge
	int drawVertex;                    // index of previously drawn vertex when drawing an edge
	                                   // -1 if not currently drawing an edge

	GLuint *gltexture;                 // array of OpenGL texture IDs

	// possible application run modes
	enum appMode {initMode = 0, designMode = 1, simMode = 2, trainMode = 3};
	appMode mode;                      // current mode

	/** Initializes the textures. */
	void initTextures();

	/**
	 * Sets the OpenGL color given by HSV coordinates.
	 * @param h the hue (0.0 <= h <= 360.0)
	 * @param s the saturation (0.0 <= s <= 1.0)
	 * @param v the value (0.0 <= v <= 1.0)
	 */
	void setHSVcolor(float h, float s, float v);
	
	/**
	 * Maps mouse coordinates to world coordinates.
	 * @param x the x coordinate of the mouse position
	 * @param y the y coordinate of the mouse position
	 */
	vector2 getPos(int x, int y);
	
	/**
	 * Rounds coordinates to integer values.
	 * @param v the vector that should be rounded
	 */
	vector2 roundPos(const vector2 &v);
	
	/**
	 * Detects if mouse is near a bridge vertex or edge and sets selVertex
	 * and selEdge according to this. Variable cursor will also be changed.
	 */
	void doMouseDetection(int x, int y);

	/**
	 * A general OpenGL initialization function. Sets all of the initial
	 * parameters. Should be called right after the OpenGL window is created.
	 */
	void initGL(int Width, int Height);

	/**
	 * The main drawing function.
	 */
	void drawGLScene();
	
	/**
	 * Handles all keyboard and mouse events.
	 */
	int handleEvents();

public:
	/**
	 * Constructor of the graphical user interface.
	 * @param w the world
	 */
	BridgeGUI(World *w): world(w) {}
	
	/**
	 * Starts the GUI.
	 * @param width the width of the window
	 * @param height the height of the window
	 * @param fullScreen if true then GUI will run in full screen mode
	 * @param bridgeFile the filename of the bridge description file
	 * @param trainFile the filename of the train description file
	 */
	void run(int width, int height, bool fullScreen, char *bridgeFile, char *trainFile);
};

#endif
