/*
 * Physically Based Simulation, SS 2005
 *
 * Markus Liechti (liechtim@student.ethz.ch)
 * Yoshimi Takano (yoshi@student.ethz.ch)
 * Stefan Wismer (swismer@student.ethz.ch)
 *
 * $Id: smoke.cpp,v 1.8 2005/06/25 20:00:04 ytakano Exp $
 */

#include "smoke.h"

Smoke::Smoke() {
	// load bitmap
	texture = SDL_LoadBMP(SMOKE);
	if (!texture) {
		cerr << "Error while loading " << SMOKE << ": " << SDL_GetError() << endl;
		exit(1);
	}

	// load mask
	mask = SDL_LoadBMP(SMOKEMASK);
	if (!mask) {
		cerr << "Error while loading " << SMOKEMASK << ": " << SDL_GetError() << endl;
		exit(1);
	}
	
	SDL_LockSurface(texture);
	SDL_LockSurface(mask);
	textureWidth = texture->w;
	textureHeight = texture->h;
	image = new unsigned char[4 * texture->w * texture->h];
	for (int i = 0; i < texture->w * texture->h; i++) {
		image[i * 4 + 0] = ((unsigned char*)(texture->pixels))[i * 3 + 2];
		image[i * 4 + 1] = ((unsigned char*)(texture->pixels))[i * 3 + 1];
		image[i * 4 + 2] = ((unsigned char*)(texture->pixels))[i * 3 + 0];
		image[i * 4 + 3] = ((unsigned char*)(mask->pixels))[i];
	}
	SDL_UnlockSurface(mask);
	SDL_UnlockSurface(texture);
}
void Smoke::clear() {
	particles.clear();
}

void Smoke::createSmoke(vector2 &start, vector2 &velocity, int n) {
	for (int i = 0; i < n; i++) {
		component mass = perturbate(SMOKE_MASS_MEAN, SMOKE_MASS_STD);
		vector2 pos = perturbate(start, SMOKE_POS_STD);
		vector2 vel = perturbate(velocity, SMOKE_VEL_LENGTH_STD, SMOKE_VEL_ANGLE_MEAN, SMOKE_VEL_ANGLE_STD);
		component lifetime = perturbate(SMOKE_LIEFTIME_MEAN, SMOKE_LIEFTIME_STD);

		particles.push_back(Particle(mass, pos, vel, lifetime));
	}
}

void Smoke::setExternalForces() {
	vector2 forceMean(SMOKE_FORCE_MEAN_X, SMOKE_FORCE_MEAN_Y);
	for (list<Particle>::iterator i = particles.begin(); i != particles.end(); i++) {
		vector2 force = perturbate(forceMean, SMOKE_FORCE_STD);
		// make force depend on the age of the particle
		i->force = force * (SMOKE_FORCE_AGE_COEFF - i->age/i->lifetime);
	}
}

void Smoke::doPhysics(component timestep) {
	setExternalForces();

	list<Particle>::iterator i = particles.begin();
	while (i != particles.end()) {
		// leap frog integration
		vector2 a = (i->force - SMOKE_DAMPING * i->velocity) / i->mass;
		i->velocity += timestep * a;
		i->position += timestep * i->velocity;
		i->age += timestep;

		// check age
		if (i->age > i->lifetime) {
			i = particles.erase(i);
		} else {
			i++;
		}
	}
}

void Smoke::draw() {
	glBegin(GL_QUADS);
	for (list<Particle>::iterator i = particles.begin(); i != particles.end(); i++) {
		glColor4f(1.0, 1.0, 1.0, 1.0 - i->age/i->lifetime);

		component displacement = 0.5 * i->age;
		component left = i->position.x - displacement;
		component right = i->position.x + displacement;
		component bottom = i->position.y - displacement;
		component top = i->position.y + displacement;
	
		glTexCoord2f(0.0, 1.0);
		glVertex3f(left, top, 2.0);
		glTexCoord2f(1.0, 1.0);
		glVertex3f(right, top, 2.0);
		glTexCoord2f(1.0, 0.0);
		glVertex3f(right, bottom, 2.0);
		glTexCoord2f(0.0, 0.0);
		glVertex3f(left, bottom, 2.0);
	}
	glEnd();
}
