/*
 * Physically Based Simulation, SS 2005
 *
 * Markus Liechti (liechtim@student.ethz.ch)
 * Yoshimi Takano (yoshi@student.ethz.ch)
 * Stefan Wismer (swismer@student.ethz.ch)
 *
 * $Id: structures.h,v 1.22 2005/06/26 18:40:26 ytakano Exp $
 */

#ifndef __structures_H__
#define __structures_H__

#include "vector2.h"

/**
 * Represents a (bridge or train) vertex.
 */
struct vertex {
	/** The possible vertex types. */
	enum vertexType {normal = 0, fixed = 1, fixedTrainStartOrEnd = 2};
	
	/** The vertex type. */
	vertexType type;
	
	/** The mass. */
	component mass;
	
	/** The initial position. */
	vector2 initialPosition;
	
	/** The current position. */
	vector2 position;
	
	/** The current velocity. */
	vector2 velocity;
	
	/** The current resulting force applied to this vertex. */
	vector2 force;
	
	/** The vertical component of the force exerted to the bridge by the train. */
	component forceCoach;

	/** Constructs an uninitialized vertex. */
	vertex() {}
	
	/**
	 * Constructs a vertex.
	 * @param x the x coordinate of the initial position
	 * @param y the y coordinate of the initial position
	 */
	vertex(component x, component y): initialPosition(x, y) {}

	/**
	 * Equality operator.
	 * @param v the vertex to compare with
	 * @return true iff the initial positions match (regardless of any other
	 * properties)
	 */
	bool operator==(const vertex &v) {
		return v.initialPosition == initialPosition; // this not: && v.type == type
	}
};

/**
 * Represents a (bridge or train) edge.
 */
struct edge  {
	/** The index of the start vertex of this edge. */
	int start;
	
	/** The index of the end vertex of this edge. */
	int end;
	
	/** True iff this edge belongs to the horizontal end-to-end road. */
	bool isRoad;
	
	/** The initial length. */
	component initialLength;
	
	/** The current length. */
	component length;
	
	/** True iff this edge is broken. */
	bool isBroken;
	
	/**
	 * Computes the relative deformation of this edge. edge::initialLength and
	 * edge::length should be set properly before any call to this method.
	 * @return the relative deformation
	 */
	component deformation() {
		return  length / initialLength - 1.0;
	}
	
	/**
	 * Equality operator.
	 * @param e the edge to compare with
	 * @return true iff the start and end indices match
	 */
	bool operator==(const edge &e) {
		return e.start == start && e.end == end;
	}
};

#endif
