/* $Id: train.cpp,v 1.6 2005/06/25 20:00:05 ytakano Exp $ */

#include <vector>
#include <iostream>
#include <fstream>
#include <algorithm>
#include "train.h"

using namespace std;

void Train::loadTextures() {
	// load bitmap
	texture = SDL_LoadBMP(filenameTexture);
	if (!texture) {
		cerr << "Error while loading " << filenameTexture << ": " << SDL_GetError() << endl;
		exit(1);
	}

	// load mask
	mask = SDL_LoadBMP(filenameMask);
	if (!mask) {
		cerr << "Error while loading " << filenameMask << ": " << SDL_GetError() << endl;
		exit(1);
	}
	
	SDL_LockSurface(texture);
	SDL_LockSurface(mask);
	textureWidth = texture->w;
	textureHeight = texture->h;
	image = new unsigned char[4 * texture->w * texture->h];
	for (int i = 0; i < texture->w * texture->h; i++) {
		image[i * 4 + 0] = ((unsigned char*)(texture->pixels))[i * 3 + 2];
		image[i * 4 + 1] = ((unsigned char*)(texture->pixels))[i * 3 + 1];
		image[i * 4 + 2] = ((unsigned char*)(texture->pixels))[i * 3 + 0];
		image[i * 4 + 3] = ((unsigned char*)(mask->pixels))[i];
	}
	SDL_UnlockSurface(mask);
	SDL_UnlockSurface(texture);
}

void Train::draw() {
	glColor3f(1.0, 1.0, 1.0);
	glBegin(GL_QUADS);
	for (unsigned int i = 0; i < coachEdges.size(); i++) {
		vector2 &vStart = coachVertices[coachEdges[i].start].position;
		vector2 &vEnd = coachVertices[coachEdges[i].end].position;
		vector2 normal = coaches[i].height * perpendicular(vEnd - vStart);
		
		glTexCoord2f(coaches[i].texRight, coaches[i].texBottom);
		glVertex3f(vStart.x, vStart.y, 1.0);
		glTexCoord2f(coaches[i].texRight, coaches[i].texTop);
		glVertex3f(vStart.x - normal.x, vStart.y - normal.y, 1.0);
		glTexCoord2f(coaches[i].texLeft, coaches[i].texTop);
		glVertex3f(vEnd.x - normal.x, vEnd.y - normal.y, 1.0);
		glTexCoord2f(coaches[i].texLeft, coaches[i].texBottom);
		glVertex3f(vEnd.x, vEnd.y, 1.0);
	}
		
	glEnd();
}

bool Train::loadFromFile(char *filename) {
	ifstream fin;
	fin.open(filename, ios::in);
	if (!fin.is_open()) {
		return false;
	}
	coaches.clear();
	
	// read texture and mask filename
	fin >> filenameTexture >> filenameMask;
	
	// read length, height and texture coordinates of each coach
	coach tmpC;
	int size;
	fin >> size;
	for (int i = 0; i < size; i++) {
		fin >> tmpC.mass >> tmpC.length >> tmpC.height;
		fin >> tmpC.texLeft >> tmpC.texTop >> tmpC.texRight >> tmpC.texBottom;
		coaches.push_back(tmpC);
	}
	
	fin >> hasSmoke;
	
	if (hasSmoke) {
		// read particle source
		fin >> smokeSource.x >> smokeSource.y;
	}
	
	fin.close();
	loadTextures();
	return true;
}

bool Train::initialize(Bridge &bridge) {
	this->bridge = &bridge;
	bridge.getRoad();
	
	if (bridge.road.size() == 0) return false;

	coachVertices.clear();
	coachEdges.clear();

	trainLength = 0;
	
	vertex tmpV;
	tmpV.velocity = vector2(0.0, 0.0);
	tmpV.position = bridge.vertices[bridge.roadStartVertex].initialPosition;
	tmpV.position.x -= 10;
	tmpV.mass = 0;
	coachVertices.push_back(tmpV);
	for (unsigned int i = 0; i < coaches.size(); i++) {
		trainLength += coaches[i].length;
		tmpV.position.x -= coaches[i].length;
		coachVertices.push_back(tmpV);
	}
	 
	edge tmpE;
	tmpE.isBroken = false;
	for (unsigned int i = 0; i < coaches.size(); i++) {
		tmpE.length = tmpE.initialLength = coaches[i].length;
		tmpE.start = i;
		tmpE.end = i + 1;
		coachEdges.push_back(tmpE);
		
		coachVertices[i].mass += 0.5 * coaches[i].mass;
		coachVertices[i + 1].mass += 0.5 * coaches[i].mass;
	}
	stopTrain = false;
	
	return true;
}

bool Train::doPhysics(component timestep) {
	if (bridge->road.size() == 0) return false;

	bridge->setExternalForces(coachVertices);

	bridge->addInternalForces(coachVertices, coachEdges, COACH_STIFFNESS, false);
	
	// adjust speed
	vector2 direction = coachVertices[0].position - coachVertices[1].position;
	direction = normalize(direction);
	if (!stopTrain) {
		if (length(coachVertices[0].velocity) < 5) {
			//speed up slowly
			coachVertices[0].force += 1 * coachVertices[0].mass * direction / timestep;
		} else {
			if (length(coachVertices[0].velocity) < 10) coachVertices[0].force += 2 * coachVertices[0].mass * direction / timestep;
		}
		if (length(coachVertices[0].velocity) > 15) {
			coachVertices[0].force -= 5 * coachVertices[0].mass * direction / timestep;
		} else {
			if (length(coachVertices[0].velocity) > 20) coachVertices[0].force -= 0.5 * length(coachVertices[0].velocity) * coachVertices[0].mass * direction / timestep;
		}
	} else {
		// stop the train
		if (length(coachVertices[0].velocity) > 1) {
			coachVertices[0].force -= 0.05 * coachVertices[0].mass * direction / timestep;
		}
	}

	vector2 oldPosition;
	for (unsigned int i = 0; i < coachVertices.size(); i++) {
		oldPosition = coachVertices[i].position;
		
		// calculate new position (without collision)
		vector2 a = (coachVertices[i].force - DAMPING * coachVertices[i].velocity) / coachVertices[i].mass;
		
		coachVertices[i].velocity += timestep * a;
		coachVertices[i].position += timestep * coachVertices[i].velocity;
		
		// detect in which segment (x-axis) we are
		int segment = -1;
		for (unsigned int j = 0; j < bridge->road.size(); j++) {
			if (bridge->edges[bridge->road[j]].isBroken) {
				stopTrain = true;
			}
			vector2 &start = bridge->getStart(bridge->road[j]);
			vector2 &end = bridge->getEnd(bridge->road[j]);
			if ((start.x <= coachVertices[i].position.x && end.x >= coachVertices[i].position.x) ||
				(end.x <= coachVertices[i].position.x && start.x >= coachVertices[i].position.x)) {
				segment = j;
				break;
			}
		}
		if (segment != -1 && bridge->edges[bridge->road[segment]].isBroken) {
			continue;
		}
		// intersection test
		if (segment == -1) { // before or after bridge
			// if coach is under ground move it up
			component ground = bridge->vertices[bridge->roadStartVertex].position.y;
			if (coachVertices[i].position.y < ground) {
				coachVertices[i].position.y = ground;
				coachVertices[i].velocity = (coachVertices[i].position - oldPosition) / timestep;
			}
			if(coachVertices[i].position.x > bridge->vertices[bridge->roadEndVertex].position.x + trainLength + 40) {
				finalize();
				return true;
			}
		} else {
			vector2 &start = bridge->getStart(bridge->road[segment]);
			vector2 &end = bridge->getEnd(bridge->road[segment]);
			if (below(coachVertices[i].position, start, end) == true) {
				// project onto bridge
				coachVertices[i].position = project(coachVertices[i].position, start, end);
				coachVertices[i].velocity = (coachVertices[i].position - oldPosition) / timestep;

				// add train mass to vertices
				for (unsigned int j = 0; j < bridge->road.size(); j++) {
					bridge->vertices[bridge->edges[bridge->road[j]].start].forceCoach = 0;
					bridge->vertices[bridge->edges[bridge->road[j]].end].forceCoach = 0;
				}
				if (end.x != start.x) {
					double factor = (coachVertices[i].position.x - start.x) / (end.x - start.x);
					bridge->vertices[bridge->edges[bridge->road[segment]].start].forceCoach = (1 - factor) * coachVertices[i].mass * GRAVITY * MASS_FACTOR;
					bridge->vertices[bridge->edges[bridge->road[segment]].end].forceCoach = factor * coachVertices[i].mass * GRAVITY * MASS_FACTOR;
				}

				// has segment changed?
				/*component rightEnd = (end.x > start.x ? end.x : start.x);
				if (coachVertices[i].position.x > rightEnd && segment+1 < (int) bridge->road.size()) {
					// project again on the new segment
					coachVertices[i].position = project(coachVertices[i].position, bridge->vertices[bridge->edges[bridge->road[segment+1]].start].position, bridge->vertices[bridge->edges[bridge->road[segment+1]].end].position);
					coachVertices[i].velocity = (coachVertices[i].position - oldPosition) / timestep;
				}*/
			}
		}
	}

	if (hasSmoke) {
		// create smoke
		vector2 smokePosition(coachVertices[0].position);

		// add displacements
		vector2 &vStart = coachVertices[1].position;
		vector2 &vEnd = coachVertices[0].position;
		vector2 normal = smokeSource.y * perpendicular(vEnd - vStart);
		smokePosition = smokePosition - smokeSource.x * normalize(vEnd - vStart) + smokeSource.y * normal;
		
		// compute velocity
		vector2 vel(coachVertices[0].velocity + 8.0 * normal);
		smoke.createSmoke(smokePosition, vel, 2);

		// perform smoke simulation
		smoke.doPhysics(timestep);
	}
	
	return false;
}

void Train::finalize() {
	bridge->road.clear();
	smoke.clear();
	bridge->puff.clear();
	for (unsigned int i = 0; i < bridge->vertices.size(); i++) {
		bridge->vertices[i].forceCoach = 0;
	}
}

void Train::cleanUp() {
	// free textures
	SDL_FreeSurface(texture);
	SDL_FreeSurface(mask);
	delete [] image;
}

