/*
 * Physically Based Simulation, SS 2005
 * 
 * Markus Liechti (liechtim@student.ethz.ch)
 * Yoshimi Takano (yoshi@student.ethz.ch)
 * Stefan Wismer (swismer@student.ethz.ch)
 *
 * $Id: train.h,v 1.7 2005/06/25 20:18:50 ytakano Exp $
 */

/** @file */
 
#ifndef __train_H__
#define __train_H__

#include <vector>
#include <cmath>
#include "opengl.h"
#include "bridge.h"
#include "structures.h"

/** The coach stiffness. */
#define COACH_STIFFNESS 20000000.0

/** The coach mass multiplication factor. */
#define MASS_FACTOR 25

/**
 * Represents a coach.
 */
struct coach {
	/** mass of the coach */
	component mass;

	/** length and height of the coach */
	component length, height;

	/** texture coordinates of the coach */
	GLfloat texLeft, texRight, texTop, texBottom;

	bool brokenAfterwards;
};

/**
 * Represents a train.
 */
class Train {
private:
	char filenameTexture[255];
	char filenameMask[255];
	SDL_Surface *texture, *mask;
	
	bool stopTrain;
	Bridge *bridge;
	
	void loadTextures();

public:
	/** The coaches. */
	vector<coach> coaches;

	/** The length of the train. */
	component trainLength;
	
	/** The vertices representing the train. */ 
	vector<vertex> coachVertices;
	
	/** The edges representing the train. */
	vector<edge> coachEdges;
	
	/** The dimensions of the texture image. */
	int textureWidth, textureHeight;
	
	/** The texture image. */
	unsigned char *image;

	/** True iff this train has smoke. */
	bool hasSmoke;

	/** The smoke particle system. */
	Smoke smoke;

	/** The source position of the smoke. */
	vector2 smokeSource;
	
	/**
	 * Draws the train.
	 */
	 void draw();
	
	/**
	 * Loads a train from a file.
	 * @return true iff loading is successful */
	bool loadFromFile(char *filename);
	
	/**
	 * Creates the train and checks if an end-to-end road can be found. Should be
	 * called before doTrain(component).
	 * @return true iff a road can be found */
	bool initialize(Bridge &bridge);
	
	/**
	 * Simulates the train.
	 * @param timestep the time step
	 * @return true iff the train has successfully passed the bridge
	 */
	bool doPhysics(component timestep);

	/**
	 * Stops the train simulation.
	 */
	void finalize();
	
	/**
	 * Cleans up after train simulation.
	 */
	void cleanUp();
};

#endif
