/*
 * Physically Based Simulation, SS 2005
 *
 * Markus Liechti (liechtim@student.ethz.ch)
 * Yoshimi Takano (yoshi@student.ethz.ch)
 * Stefan Wismer (swismer@student.ethz.ch)
 *
 * $Id: vector2.cpp,v 1.10 2005/06/26 16:36:37 ytakano Exp $
 */

#include "vector2.h"
#include "structures.h"

void vector2::operator+=(const component c) {
	x += c;
	y += c;
}

void vector2::operator+=(const vector2 &u) {
	x += u.x;
	y += u.y;
}

void vector2::operator-=(const component c) {
	x -= c;
	y -= c;
}

void vector2::operator-=(const vector2 &u) {
	x -= u.x;
	y -= u.y;
}

void vector2::operator*=(const component c) {
	x *= c;
	y *= c;
}

void vector2::operator*=(const vector2 &u) {
	x *= u.x;
	y *= u.y;
}

void vector2::operator/=(const component c) {
	x /= c;
	y /= c;
}

void vector2::operator/=(const vector2 &u) {
	x /= u.x;
	y /= u.y;
}

vector2::operator vertex() {
	return vertex(x, y);
}

void vector2::operator=(const vertex &v) {
	x = v.initialPosition.x;
	y = v.initialPosition.y;
}

vector2 operator+(const vector2 &u, const vector2 &v) {
	return vector2(u.x + v.x, u.y + v.y);
}

vector2 operator+(const vector2 &u, const component c) {
	return vector2(u.x + c, u.y + c);
}

vector2 operator+(const component c, const vector2 &u) {
	return vector2(u.x + c, u.y + c);
}

vector2 operator-(const vector2 &u, const vector2 &v) {
	return vector2(u.x - v.x, u.y - v.y);
}

vector2 operator-(const vector2 &u, const component c) {
	return vector2(u.x - c, u.y - c);
}

vector2 operator-(const component c, const vector2 &u) {
	return vector2(u.x - c, u.y - c);
}

vector2 operator-(const vector2 &u) {
	return vector2(-u.x, -u.y);
}

component operator*(const vector2 &u, const vector2 &v) {
	return u.x * v.x + u.y * v.y;
}

component operator%(const vector2 &u, const vector2 &v) {
	return u.x * v.y - u.y * v.x;
}

vector2 operator*(const component c, const vector2 &u) {
	return vector2(u.x * c, u.y * c);
}

vector2 operator*(const vector2 &u, const component c) {
	return vector2(u.x * c, u.y * c);
}

vector2 operator/(const vector2 &u, const vector2 &v) {
	return vector2(u.x / v.x, u.y / v.y);
}

vector2 operator/(const vector2 &u, const component c) {
	return vector2(u.x / c, u.y / c);
}

vector2 operator/(const component c, const vector2 &u) {
	return vector2(c / u.x, c / u.y);
}

bool operator==(const vector2 &u, const vector2 &v) {
	return u.x == v.x && u.y == v.y;
}

bool operator!=(const vector2 &u, const vector2 &v) {
	return !(u == v);
}

component length(const vector2 &u) {
	return sqrt(u.x * u.x + u.y * u.y);
}

vector2 normalize(const vector2 &u) {
	return u/length(u);
}

vector2 signSquare(vector2 &v) {
	return vector2(abs(v.x) * v.x, abs(v.y) * v.y);
}

vector2 project(vector2 &u, vector2 &vStart, vector2 &vEnd) {
	if (vStart == vEnd) return u;
	component p = ((vEnd - vStart) * (u - vStart)) / length(vEnd - vStart);
	return vStart + p * ((vEnd - vStart) / length(vEnd - vStart));
}

bool below(vector2 &u, vector2 &vStart, vector2 &vEnd) {
	if (vStart == vEnd) return false;
	vector2 normal = vector2(vStart.y - vEnd.y, vEnd.x - vStart.x);
	// project point u onto normal
	component p = (normal * (u - vStart)) / length(normal);
	
	if (vStart.x < vEnd.x) {
		if (p < -1) return false;
		return (p < 0);
	} else {
		if (p > 1) return false;
		return (p > 0);
	}
}

vector2 perpendicular(const vector2 &u) {
	return normalize(vector2(-u.y, u.x));
}

vector2 rotate(vector2 &u, component theta) {
	component s = sin(theta), c = cos(theta);
	return vector2(c*u.x - s*u.y, s*u.x + c*u.y);
}

component perturbate(component mu, component sigma) {
	#define PI 3.1415927
	return mu + sigma * sqrt(-2.0*log((rand()+1.0)/RAND_MAX))*sin(2.0*PI*rand()/RAND_MAX);
}

vector2 perturbate(vector2 &mu, component sigma) {
	component x = perturbate(mu.x, sigma);
	component y = perturbate(mu.y, sigma);
	return vector2(x, y);
}

vector2 perturbate(vector2 &u, component sigma1, component theta, component sigma2) {
	vector2 v = perturbate(length(u), sigma1) * normalize(u);
	component t = perturbate(theta, sigma2);
	return rotate(v, t);
}

ostream &operator<<(ostream &os, const vector2 &u) {
	os << "(" << u.x << ", " << u.y << ")";
	return os;
}
