/*
 * Physically Based Simulation, SS 2005
 *
 * Markus Liechti (liechtim@student.ethz.ch)
 * Yoshimi Takano (yoshi@student.ethz.ch)
 * Stefan Wismer (swismer@student.ethz.ch)
 *
 * $Id: vector2.h,v 1.12 2005/06/26 18:40:26 ytakano Exp $
 */

/** @file */

#ifndef _VECTOR2_H_
#define _VECTOR2_H_

#include <iostream>
#include <cmath>

using namespace std;

struct vertex;

/** The type of the vector components. */
typedef double component;

/**
 * Represents a 2D vector.
 */
struct vector2 {
	/** The x coordinate. */
	component x;

	/** The y coordinate. */
	component y;

	/** Constructs an uninitialized vector. */
	vector2() {}

	/** Constructs a vector.
	 * @param xp the x coordinate
	 * @param yp the y coordinate
	 */
	vector2(const component xp, const component yp) : x(xp), y(yp) {}

	/** Per-component addition of a scalar. */
	void operator+=(const component c);
	
	/** Component-wise addition. */
	void operator+=(const vector2 &u);
	
	/** Per-component subtraction of a scalar. */
	void operator-=(const component c);
	
	/** Component-wise subtraction. */
	void operator-=(const vector2 &u);
	
	/** Multiplication with a scalar. */
	void operator*=(const component c);
	
	/** Component-wise multiplication. */
	void operator*=(const vector2 &u);
	
	/** Division by a scalar. */
	void operator/=(const component c);
	
	/** Component-wise division. */
	void operator/=(const vector2 &u);
	
	/** Converts this vector to a vertex with the same coordinates. */
	operator vertex();
	
	/** Assignment operator. */
	void operator=(const vertex &v);
};

/** Component-wise addition. */
vector2 operator+(const vector2 &u, const vector2 &v);

/** Per-component addition of a scalar. */
vector2 operator+(const vector2 &u, const component c);

/** Per-component addition of a scalar. */
vector2 operator+(const component c, const vector2 &u);

/** Component-wise subtraction. */
vector2 operator-(const vector2 &u, const vector2 &v);

/** Per-component subtraction of a scalar. */
vector2 operator-(const vector2 &u, const component c);

/** Per-component subtraction from a scalar. */
vector2 operator-(const component c, const vector2 &u);

/** Negation. */
vector2 operator-(const vector2 &u);

/** Scalar product. */
component operator*(const vector2 &u, const vector2 &v);

/** Multiplication with a scalar. */
vector2 operator*(const component c, const vector2 &u);

/** Multiplication with a scalar. */
vector2 operator*(const vector2 &u, const component c);

/** Component-wise division. */
vector2 operator/(const vector2 &u, const vector2 &v);

/** Division by a scalar. */
vector2 operator/(const vector2 &u, const component c);

/** Component-wise scalar division. */
vector2 operator/(const component c, const vector2 &u);

/** 2D cross product. */
component operator%(const vector2 &u, const vector2 &v);

/** Equality test. */
bool operator==(const vector2 &u, const vector2 &v);

/** Inequality test. */
bool operator!=(const vector2 &u, const vector2 &v);

/** Euclidean length. */
component length(const vector2 &u);

/** Normalization. */
vector2 normalize(const vector2 &u);

/**
 * Computes the component-wise square of a given vector retaining the sign.
 * @param v the vector
 * @return the component-wise square of the vector with the sign retained
 */
vector2 signSquare(vector2 &v);

/**
 * Projects point onto a line.
 * @param u the point to be projected
 * @param vStart the start point of the line
 * @param vEnd the end point of the line
 * @return the projected point
 */
vector2 project(vector2 &u, vector2 &vStart, vector2 &vEnd);

/**
 * Checks if a given point is below a given segment.
 * @param u the point
 * @param vStart start point of the segment
 * @param vEnd the end point of the segment
 * @return true iff the point is below the segment
 */
bool below(vector2 &u, vector2 &vStart, vector2 &vEnd);

/**
 * Returns the normalized perpendicular of a given vector.
 * @param u the vector of which the perpendicular is computed
 * @return the normalized perpendicular
 */
vector2 perpendicular(const vector2 &u);

/**
 * Rotates a given vector around the origin by a given angle.
 * @param u the vector to be rotated
 * @param theta the rotation angle
 * @return the rotated vector
 */
vector2 rotate(vector2 &u, component theta);

/**
 * Computes an approximated sample of a normal distribution.
 * @param mu the mean
 * @param sigma the standard deviation
 * @return an approximated sample of the normal distribution N(mu, sigma)
 */
component perturbate(component mu, component sigma);

/**
 * Perturbates a given vector approximately according to a normal distribution.
 * Equivalent to applying perturbate(mu.x, sigma) and perturbate(mu.y, sigma).
 * @param mu the mean vector
 * @param sigma the standard deviation
 * @return the approximatly normally perturbated vector
 * @sa perturbate(component, component)
 */
vector2 perturbate(vector2 &mu, component sigma);

/**
 * Perturbates a given vector approximately according to a normal distribution.
 * The following steps are performed:
 * - perturbate the length with a given standard deviation
 * - rotate the vector around an angle with given mean and standard deviation
 * @param u the vector to be perturbated
 * @param sigma1 the standard deviation for the length perturbation
 * @param theta the mean roation angle
 * @param sigma2 the standard deviation of the rotation angle
 * @return the perturbated vector
 * @sa perturbate(component, component)
 */
vector2 perturbate(vector2 &u, component sigma1, component theta, component sigma2);

/**
 * Writes a vector to a given output stream.
 * @param os the output stream
 * @param u the vector
 * @return the written-to output stream
 */
ostream &operator<<(ostream &os, const vector2 &u);

#endif
